#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>
#include <string.h>

#define VCC 0x80
#define RESET 0x40
#define SDO_PULLDOWN 0x08
#define SDO 0x08
#define SCL 0x04
#define SII 0x02
#define SDI 0x01

int parportFD;

void delay(int us) {
//	usleep(us);
//	sleep(1);
//	usleep(10000);
//	usleep(1);
}

int setupPort(char *device) {
	parportFD = open(device, O_RDWR);
	if(parportFD < 0) {
		perror(device);
		exit(1);
	}
	
	if(ioctl(parportFD, PPCLAIM)) {
		perror("ioctl(PPCLAIM)");
		exit(2);
	}

	return parportFD;
}

void closePort() {
	if(ioctl(parportFD, PPRELEASE)) {
		perror("ioctl(PPRELEASE)");
		exit(2);
	}
	
	if(close(parportFD)) {
		perror("close()");
		exit(1);
	}
}

void setData(int d) {
	if(ioctl(parportFD, PPWDATA, &d)) {
		perror("ioctl(PPWDATA)");
		exit(3);
	}
}

int readSDO() {
	int s;

	if(ioctl(parportFD, PPRSTATUS, &s)) {
		perror("ioctl(PPRSTATUS)");
		exit(3);
	}

//	printf("S: %d\n", ((s & SDO)?(1):(0)));

	return ((s & SDO)?(1):(0));
}

void powerUp() {
	// start from off
	setData(0);
	usleep(10000);

	// pull down SDO
	setData(SDO_PULLDOWN);
	usleep(50);

	// bring up VCC and RESET (12 Volts)
	setData(VCC | SDO_PULLDOWN);
	setData(VCC | RESET | SDO_PULLDOWN);
	usleep(10000);
	
	// release SDO pull-down
	setData(VCC | RESET);
	usleep(200);
}

void powerDown() {
	setData(0);
}

unsigned char sendDI(unsigned char data, unsigned char instruction) {
	int i, d, sdo;
	unsigned char out = 0;

	d = VCC | RESET;
	for(i = 0; i < 8; i++) {
		d |= SCL;
		setData(d);
		delay(1);
		d &= ~SCL;
		setData(d);
		delay(1);
		d &= ~(SDI | SII);
		d |= ((data & (1 << (7 - i)))?(SDI):(0)) | ((instruction & (1 << (7 - i)))?(SII):(0));
		setData(d);
		delay(1);
		sdo = readSDO();
		out |= (((sdo)?(1):(0)) << (7 - i));
		delay(1);
	}
	for(i = 0; i < 3; i++) {
		d |= SCL;
		setData(d);
		delay(1);
		d = VCC | RESET;
		setData(d);
		delay(1);
	}
	
//	printf("0x%02X 0x%02X 0x%02X\n", data, instruction, out);

	return out;
}

char readSignature(int addr) {
	sendDI(0x08, 0x4C);
	sendDI(addr & 0x03, 0x0C);
	sendDI(0x00, 0x68);
	return sendDI(0x00, 0x6C);
}

char readCalibration() {
	sendDI(0x08, 0x4C);
	sendDI(0x00, 0x0C);
	sendDI(0x00, 0x78);
	return sendDI(0x00, 0x7C);
}

char readFuseLow() {
	sendDI(0x04, 0x4C);
	sendDI(0x00, 0x68);
	return sendDI(0x00, 0x6C);
}

char readFuseHigh() {
	sendDI(0x04, 0x4C);
	sendDI(0x00, 0x7A);
	return sendDI(0x00, 0x7E);
}

char readLock() {
	sendDI(0x04, 0x4C);
	sendDI(0x00, 0x78);
	return sendDI(0x00, 0x7A);
}

void chipErase() {
	sendDI(0x80, 0x4C);
	sendDI(0x00, 0x64);
	sendDI(0x00, 0x6C);
	printf("erase waiting... ");
	while(!readSDO()) {
		usleep(1);
		printf("#");
	}
	printf(" done\n");
}

void writeFuseLow(unsigned char fuseBits) {
	sendDI(0x40, 0x4C);
	sendDI(fuseBits, 0x2C);
	sendDI(0x00, 0x64);
	sendDI(0x00, 0x6C);
	printf("write waiting... ");
	while(!readSDO()) {
		usleep(1);
		printf("#");
	}
	printf(" done\n");
}

void writeFuseHigh(unsigned char fuseBits) {
	sendDI(0x40, 0x4C);
	sendDI(fuseBits, 0x2C);
	sendDI(0x00, 0x74);
	sendDI(0x00, 0x7C);
	printf("write waiting... ");
	while(!readSDO()) {
		usleep(1);
		printf("#");
	}
	printf(" done\n");
}

void nop() {
	sendDI(0x00, 0x4C);
}

int main(int argc, char **argv) {
	int i;
	unsigned char c[3];
	
	setupPort("/dev/parport0");
	powerUp();

//	for(i = 0; i < 5; i++) sendDI(0x00, 0x4C);

	for(i = 0; i < 3; i++) {
		c[i] = readSignature(i);
		printf("sig %d: 0x%02X\n", i, c[i] & 0xFF);
	}
	
	if(c[0] != 0x1E && c[1] != 0x90 && c[2] != 0x07) {
		printf("Not an ATtiny13!\n");
		powerDown();
		closePort();
		exit(4);
	}

	printf("cal: 0x%02X\n", readCalibration() & 0xFF);
	printf("fuse low: 0x%02X\n", readFuseLow() & 0xFF);
	printf("fuse high: 0x%02X\n", readFuseHigh() & 0xFF);
	printf("lock: 0x%02X\n", readLock() & 0xFF);

	if(1) {
		chipErase();
		writeFuseLow(0x6A);
		writeFuseHigh(0xFF);
	}

	powerDown();
	closePort();

	return 0;
}
